// SPDX-License-Identifier: GPL-2.0
// Copyright (C) 2021 Jakub Horak
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/wdt.h>
#include <avr/eeprom.h>
#include <avr/power.h>
#include <util/delay.h>
#include <stdio.h>
#include <string.h>
#include "config.h"
#include "lib328/defs.h"
#include "lib328/lgt8f.h"
#include "lib328/utils.h"
#include "lib328/buttons.h"
#include "lib328/serial.h"
#include "lib328/print.h"
#include "lib328/timer.h"
#include "lib328/pins.h"
#define NO_TIMER
#include "lib328/millis.h"
#include "lib328/i2c.h"
#define F_SCL 50000

enum {
	DS1307_ADDR = 0x68,
};

enum {
	TimerClock = 1,
	PinLedSelect0 = 2,
	PinButtonA = 15,
	PinButtonB = 14,
	PinButtonOk = 13,
	PinShiftDs = 21,
	PinShiftSt = 20,
	PinShiftSh = 17,
};

#include "led595.h"

typedef struct Ui Ui;
struct Ui {
	char diff;
	byte ok;
};

static volatile Ui ui;
static volatile byte count;

static Ui
grab_ui(void)
{
	Ui ui_copy;
	ATOMIC_BLOCK(ATOMIC_RESTORESTATE) {
		ui_copy = ui;
		ui.diff = 0;
		ui.ok = 0;
	}
	return ui_copy;
}

static void
setup_pins(void)
{
	pin_mode(PinButtonA, INPUT_PULLUP);
	pin_mode(PinButtonB, INPUT_PULLUP);
	pin_mode(PinButtonOk, INPUT_PULLUP);
}


ISR(TIMER1_COMPA_vect)
{
	static byte a_hist, b_hist, ok_hist;

	tick_world_clock();
	count += 1;

	scan_but(&a_hist, pin_read(PinButtonA));
	scan_but(&b_hist, pin_read(PinButtonB));
	ui.diff += decode_enc(a_hist, b_hist);

	if ((count & 4) == 0) {
		scan_but(&ok_hist, pin_read(PinButtonOk));
		ui.ok |= ok_hist == 0x80;
	}

	display_led_tick();
}

static inline void
start_timer(void)
{
	timer_set_mode(TimerClock, TIMER_CTC, PRESCALER_64X);
	timer_output(TimerClock, TIMER_PIN_A, 249);
	timer_enable_irq(TimerClock, TIMER_IRQ_A);
}


typedef struct Time Time;

struct Time {
	byte hh, mm, ss, dow;
	int8_t temp;
};

#define from_bcd(x) (((x) & 0xf) + (((x) >> 4) * 10))
#define to_bcd(x) (((x) % 10) | (((x) / 10) << 4))

static char
i2c_write_buf(byte i2c_addr, byte addr, byte *buf, byte len)
{
	char err;
	err = i2c_start(i2c_addr, TW_WRITE);
	if (err != 0)
		return err;
	err = i2c_send_byte(addr);
	if (err != 0)
		return err;
	for (byte i = 0; i < len; i++) {
		err = i2c_send_byte(buf[i]);
		if (err != 0)
			return err;
	}
	err = i2c_stop();
	return err;
}

static char
i2c_read_buf(byte i2c_addr, byte addr, byte *buf, byte len)
{
	char err;
	err = i2c_start(i2c_addr, TW_WRITE);
	if (err != 0)
		return err;
	err = i2c_send_byte(addr);
	if (err != 0)
		return err;
	err = i2c_start(i2c_addr, TW_READ);
	if (err != 0)
		return err;
	for (byte i = 0; i < len; i++) {
		err = i2c_receive_byte(&buf[i], i + 1 != len);
		if (err != 0)
			return err;
	}
	err = i2c_stop();
	return err;
}


static char
set_time(Time *time)
{
	byte data[4] = {
		to_bcd(time->ss),
		to_bcd(time->mm),
		to_bcd(time->hh),
		time->dow,
	};
	byte mode_24hours = 0x40;
	char err;
	err = i2c_write_buf(DS1307_ADDR, 2, &mode_24hours, 1);
	if (err < 0)
		return err;
#if 0
	byte control_reg = 0;
	err = i2c_write_buf(DS1307_ADDR, 0x0e, &control_reg, 1);
	if (err < 0)
		return err;
#endif
	return i2c_write_buf(DS1307_ADDR, 0, data, sizeof(data));
}

enum {
	RegSs = 0,
	RegMm = 1,
	RegHh = 2,
	RegDow = 3,
	RegTemp = 0x11,
};

static char
get_time(Time *time)
{
	char err;
	byte data[4];
	err = i2c_read_buf(DS1307_ADDR, RegSs, data, sizeof(data));
	if (err != 0)
		return err;
	time->dow = data[3];
	time->hh = from_bcd(data[2] & 0x3f);
	time->mm = from_bcd(data[1]);
	time->ss = from_bcd(data[0]);
	err = i2c_read_buf(DS1307_ADDR, RegTemp, (void *)&time->temp, 1);
	if (err != 0)
		return err;
	return 0;
}



static const char *day_names[] = { "brb", "Pon", "Ute", "Str", "Ctv", "Pat", "Sob", "Ned" };

static inline void print_pair(byte leading_zeros, byte x, byte show, byte sep)
{
	byte a = x / 10;
	byte b = x % 10;
	if (show) {
		if (a == 0 && !leading_zeros)
			xputc(' ');
		else
			xputc('0' + a);
		xputc('0' + b);
	} else {
		xputc(' ');
		xputc(' ');
	}
	xputc(sep);
}

int
main(void)
{
	wdt_disable();
	power_usart0_enable();
	lgt8f328p_compat();

	serial_start(115200);
	serial_stdio_init();

	setup_pins();
	start_timer();
	i2c_init(F_SCL);
	i2c_set_sw_pullups();
	led_display_setup();

	sei();

	{
		Time t = { .hh = 11, .mm = 59, .ss = 59, .dow = 1 };
		char err = set_time(&t);
		if (err != 0)
			printf("set time error: %d\n", err);
	}

	Time t = {};
	byte mode =0 ;
	for (;;) {
		Ui ui = grab_ui();
		uint32_t now = millis();

		if (ui.ok) {
			mode++;
			if (mode > 4) {
				mode = 0;
				char err = set_time(&t);
				if (err != 0)
					printf("set time error: %d\n", err);
			}
		}
		if (mode == 0) {
			char err = get_time(&t);
			if (err != 0)
				printf("get time error: %d\n", err);
		}
		if (mode != 0) {
			if (mode == 1)
				adjust(&t.hh, ui.diff, 0, 23);
			if (mode == 2)
				adjust(&t.mm, ui.diff, 0, 59);
			if (mode == 3)
				adjust(&t.ss, ui.diff, 0, 59);
			if (mode == 4)
				adjust(&t.dow, ui.diff, 1, 7);
		}
		byte show = (now >> 7) & 1;
		xcls();
		print_pair(0, t.hh, mode != 1 || show, '.');
		print_pair(1, t.mm, mode != 2 || show, '.');
		print_pair(1, t.ss, mode != 3 || show, ' ');
		xputs((mode != 4 || show) ? day_names[t.dow] : "   ");
		xpresent();
		_delay_ms(10);
	}

	return 0;
}
