// SPDX-License-Identifier: GPL-2.0
// Copyright (C) 2021 Jakub Horak
/*
 * NOTE: I2C needs _external_ pullups, at least 4.7K
 * The internal pullups (INPUT_PULLUP) are too weak.
 */
#include <util/twi.h>

#define I2C_TIMEOUT_10US 100

static byte i2c_stat;

enum {
	ACK = 0,
	NACK,
};

enum {
	I2C_START_FAIL = -1,
	I2C_SLA_FAIL = -2,
	I2C_RECV_FAIL = -3,
	I2C_SEND_FAIL = -4,
	I2C_STOP_TIMEOUT = -5,
	I2C_NACK_TIMEOUT = -6,
};

// NOTE: MINIMUM FREQUENCY is about 30.5 KHz
static void
i2c_init(uint32_t f_scl)
{
	/* SCL_freq = CPU_freq / (16 + 2*TWBR*Prescaler) */
	/* TWBR = (CPU_freq / SCL_Freq - 16) / (2 * Prescaler) */
	 
	/* set prescaler = 1 */
	TWSR = 0;
	/* set divider to F_SCL */
	TWBR = F_CPU / f_scl / 2 - 8;
	/* enable */
	TWCR = _BV(TWEN);
}

static void
i2c_set_sw_pullups(void)
{
	pin_mode(I2C_SCL_PIN, INPUT_PULLUP);
	pin_mode(I2C_SDA_PIN, INPUT_PULLUP);
}

static inline byte
i2c_wait(void)
{
	byte time_left = I2C_TIMEOUT_10US;
	while (!(TWCR & _BV(TWINT))) {
		if (!--time_left) {
			i2c_stat = 7;
			return i2c_stat;
		}
		_delay_us(10);
	}

	i2c_stat = TWSR & 0xf8;
	return i2c_stat;
}

static inline void
i2c_continue(byte ack)
{
	byte cr = _BV(TWINT) | _BV(TWEN);
	if (ack)
		cr |= _BV(TWEA);
	TWCR = cr;
}

/* mode is either: TW_READ or TW_WRITE */
static char
i2c_start(byte address, byte mode)
{
	/* step: start condition */
	TWCR |= _BV(TWEN) | _BV(TWSTA) | _BV(TWINT);
	i2c_wait();
	switch (i2c_stat) {
		case TW_START:
		case TW_REP_START:
			break;
		default:
			TWCR = 0;
			return I2C_START_FAIL;
	}

	/* step: write SLA+RW */
	TWDR = (address << 1) | mode;
	i2c_continue(1);
	i2c_wait();
	switch (i2c_stat) {
		case TW_MR_SLA_ACK:
		case TW_MT_SLA_ACK:
			return ACK;
		case TW_MR_SLA_NACK:
		case TW_MT_SLA_NACK:
			TWCR = 0;
			return NACK;
		default:
			TWCR = 0;
			return I2C_SLA_FAIL;
	}
}

#define i2c_start_transmit(a) i2c_start(a, TW_WRITE)
#define i2c_start_receive(a) i2c_start(a, TW_READ)

static inline char
i2c_receive_byte(byte *x, byte reply_ack)
{
	if (reply_ack) {
		i2c_continue(1);
		if (i2c_wait() != TW_MR_DATA_ACK)
			return I2C_RECV_FAIL;
	} else {
		i2c_continue(0);
		if (i2c_wait() != TW_MR_DATA_NACK)
			return I2C_RECV_FAIL;
	}
	*x = TWDR;
	return 0;
}

static inline char
i2c_send_byte(byte x)
{
	/* step: xmit data */
	TWDR = x;
	i2c_continue(1);
	i2c_wait();
	switch (i2c_stat) {
		case TW_MT_DATA_ACK:
			return ACK;
		case TW_MT_DATA_NACK:
			return NACK;
		default:
			return I2C_SEND_FAIL;
	}
}

static char
i2c_stop(void)
{
	byte time_left = I2C_TIMEOUT_10US;
	/* step: stop condition */
	TWCR = _BV(TWEN) | _BV(TWINT) | _BV(TWSTO);
	/* wait for it to clear */
	while (TWCR & _BV(TWSTO)) {
		if (!--time_left) {
			TWCR = 0;
			return I2C_STOP_TIMEOUT;
		}
		_delay_us(10);
	}
	TWCR = 0;
	return 0;
}
